/** \file
*   MIDI send on bar.
*   Send MIDI note events at start of next bar. 
*/
                              

//metadata
string name="MIDI send on bar";
string description="Send MIDI note events at start of next bar";

// parameters definition

array<string> inputParametersNames={"Duration"};
array<double> inputParameters(inputParametersNames.length,0);

array<double> inputParametersMin={};
array<double> inputParametersMax={};
array<double> inputParametersDefault={.5};
array<int>    inputParametersSteps={};


// internal variables
MidiEvent   noteOnEvent;
MidiEvent   noteOffEvent;
bool        playing=false;
bool		newNoteOn=false;
bool		newNoteOff=false;

// temp MIDI event

double  duration=0;

//utils
double quarterNotesToSamples(double position,double bpm)
{
    return position*60.0*sampleRate/bpm;
}
double samplesToQuarterNotes(double samples,double bpm)
{
    return samples*bpm/(60.0*sampleRate);
}

// filter interface starts here---------------------
void initialize()
{
    // initialize our events
    MidiEventUtils::setType(noteOnEvent,kMidiNoteOn);
    MidiEventUtils::setNoteVelocity(noteOnEvent,127);
    noteOffEvent=noteOnEvent;
    MidiEventUtils::setType(noteOffEvent,kMidiNoteOff);
}

void processBlock(BlockData& data)
{
    // check if was playing
    bool wasPlaying=playing;

    // cannot run if no transport available
    if(@data.transport!=null)
    {
		
		//check for all new events
		for(uint i=0;i<data.inputMidiEvents.length;i++)
		{
			// get the newest note on event and store it
			MidiEventType type=MidiEventUtils::getType(data.inputMidiEvents[i]);
			int8 velocity=MidiEventUtils::getNoteVelocity(data.inputMidiEvents[i]);
			if(type==kMidiNoteOn && velocity>0) // 0 is often used for note offs
			{
				noteOnEvent=data.inputMidiEvents[i];
				newNoteOn=true;
			}
		}
		
		
		//
        // update note value and send note offs if necessary (cancel runnning notes if changed)

        // send note off on play stop
        playing=data.transport.isPlaying;
        if(!playing && wasPlaying)
        {
            noteOffEvent.timeStamp=0;
            MidiEventUtils::setNote(noteOffEvent, MidiEventUtils::getNote(noteOnEvent));
            data.outputMidiEvents.push(noteOffEvent);
        }

        // send events when playing
        if(playing && data.transport.timeSigBottom!=0)
        { 
            //convert all values to match current time signature bottom value
            double timeSigBottom=double(data.transport.timeSigBottom);
            double beginPosition=data.transport.positionInQuarterNotes*timeSigBottom/4.0;
            double endPosition=(data.transport.positionInQuarterNotes+samplesToQuarterNotes(data.samplesToProcess,data.transport.bpm))*timeSigBottom/4.0;
            double downBeat=(data.transport.currentMeasureDownBeat*timeSigBottom)/4.0;

            // check where integer position values stand in the buffer
            double beatPosition=floor(beginPosition);
            while(beatPosition<=endPosition)
            {
                int currentPositionInMeasure=int(beatPosition-downBeat+.5);
                bool firstBeat=(currentPositionInMeasure%data.transport.timeSigTop)==0;

                // manage beat note on
				
					if(newNoteOn && firstBeat && beatPosition>=beginPosition && firstBeat)
					{
						// we have a beat => send midi note on event
						noteOnEvent.timeStamp=quarterNotesToSamples((beatPosition-beginPosition)*4.0/timeSigBottom,data.transport.bpm);

						data.outputMidiEvents.push(noteOnEvent);
						newNoteOn = false;
						newNoteOff = true;
					}

					// manage beat note off
					if(newNoteOff && firstBeat && beatPosition+duration<=endPosition && beatPosition+duration>=beginPosition)
					{
						// we have a beat+duration => send midi note off event

						MidiEventUtils::setNote(noteOffEvent,MidiEventUtils::getNote(noteOnEvent));
						MidiEventUtils::setNoteVelocity(noteOffEvent,MidiEventUtils::getNoteVelocity(noteOnEvent));

						noteOffEvent.timeStamp=quarterNotesToSamples((beatPosition+duration-beginPosition)*4.0/timeSigBottom,data.transport.bpm);
						data.outputMidiEvents.push(noteOffEvent);
						newNoteOff = false;
					} 
					
				
				beatPosition++;
            }
        }
    }
}

void updateInputParametersForBlock(const TransportInfo@ )
{

    duration=inputParameters[0];
}

/** 
*  \file Midi.hxx
*  Midi utilities library for angelscript
* 
*  Created by Blue Cat Audio <services@bluecataudio.com>
*  Copyright 2011-2017 Blue Cat Audio. All rights reserved.
*
*/

/** Enumeration of Midi Events types currently supported by this library.
*
*/
enum MidiEventType
{
    kMidiNoteOff, ///< Note Off Event
    kMidiNoteOn, ///< Note On Event
    kMidiControlChange, ///< Control Change Event (CC)
    kMidiProgramChange, ///< Program Change Event
    kMidiPitchWheel,  ///< Pitch Wheel Event
    kMidiNoteAfterTouch,  ///< Note after touch
    kMidiChannelAfterTouch, ///< Channel after touch
    kUnknown ///< Other Events
};


/** Utility functions to handle Midi events.
*
*/
namespace MidiEventUtils
{
    /** Retrieves the type of the Midi event.
    *
    */
    MidiEventType getType(const MidiEvent& evt)
    {
        MidiEventType type=kUnknown;
        switch(evt.byte0 & 0xF0)
        {
        case 0x80:
            type=kMidiNoteOff;
            break;
        case 0x90:
            type=kMidiNoteOn;
            break;
        case 0xB0:
            type=kMidiControlChange;
            break;
        case 0xC0:
            type=kMidiProgramChange;
            break;
        case 0xE0:
            type=kMidiPitchWheel;
            break;
        case 0xA0:
            type=kMidiNoteAfterTouch;
            break;
        case 0xD0:
            type=kMidiChannelAfterTouch;
            break;
        }
        return type;
    }

    /** Set the type of the Midi event.
    *  @see MidiEventType enum for supported Midi events.
    */
    void setType(MidiEvent& evt,MidiEventType type)
    {
        switch(type)
        {
        case kMidiNoteOff:
            evt.byte0=0x80|(evt.byte0 & 0x0F);
            break;
        case kMidiNoteOn:
            evt.byte0=0x90|(evt.byte0 & 0x0F);
            break;
        case kMidiControlChange:
            evt.byte0=0xB0|(evt.byte0 & 0x0F);
            break;
        case kMidiProgramChange:
            evt.byte0=0xC0|(evt.byte0 & 0x0F);
            break;
        case kMidiPitchWheel:
            evt.byte0=0xE0|(evt.byte0 & 0x0F);
            break;
        case kMidiNoteAfterTouch:
            evt.byte0=0xA0|(evt.byte0 & 0x0F);
            break;
        case kMidiChannelAfterTouch:
            evt.byte0=0xD0|(evt.byte0 & 0x0F);
            break;
        }
    }

    /** Get the channel number of the event (1-16).
    *
    */
    uint8 getChannel(const MidiEvent& evt)
    {
        return (evt.byte0 & 0x0F)+1;
    }

    /** Set the channel number for the event (1-16).
    *
    */
    void setChannel(MidiEvent& evt, uint8 ch)
    {
        if(ch!=0)
            evt.byte0=(evt.byte0&0xF0)|((ch-1)&0x0F);
    }

    /** For a note event, retrieves the Midi note for the event (0-127).
    *
    */
    uint8 getNote(const MidiEvent& evt)
    {
        return evt.byte1&0x7F;
    }

    /** For a note event, sets the Midi note for the event (0-127).
    *
    */
    void setNote(MidiEvent& evt, uint8 note)
    {
        evt.byte1=(note&0x7F);
    }

    /** For a note event, retrieves the velocity for the event (0-127).
    *
    */
    uint8 getNoteVelocity(const MidiEvent& evt)
    {
        return evt.byte2 & 0x7F;
    }

    /** For a note event, sets the velocity for the event (0-127).
    *
    */
    void setNoteVelocity(MidiEvent& evt, uint8 velocity)
    {
        evt.byte2=velocity&0x7F;
    }

    /** For a CC (Control Change) event, gets the control number (0-127).
    *
    */
    uint8 getCCNumber(const MidiEvent& evt)
    {
        return evt.byte1 & 0x7F;
    }

    /** For a CC (Control Change) event, sets the control number (0-127).
    *
    */
    void setCCNumber(MidiEvent& evt,uint8 nb)
    {
        evt.byte1=nb&0x7F;
    }

    /** For a CC (Control Change) event, gets the control value (0-127).
    *
    */
    uint8 getCCValue(const MidiEvent& evt)
    {
        return evt.byte2 & 0x7F;	
    }

    /** For a CC (Control Change) event, sets the control value (0-127).
    *
    */
    void setCCValue(MidiEvent& evt,uint8 value)
    {
        evt.byte2=value & 0x7F;
    }

    /** For a Program Change event, gets the program number (0-127).
    *
    */
    uint8 getProgram(const MidiEvent& evt)
    {
        return evt.byte1&0x7F;
    }

    /** For a Program Change event, sets the program number (0-127).
    *
    */
    void setProgram(MidiEvent& evt, uint8 prog)
    {
        evt.byte1=prog&0x7F;
    }    

    /** For a pitch Wheel event, gets the pitch value (-8192 to +8192).
    *
    */
    int getPitchWheelValue(const MidiEvent& evt)
    {
        return (evt.byte2 & 0x7F)*128+(evt.byte1 & 0x7F)-64*128;	
    }

    /** For a pitch Wheel event, sets the pitch value (-8192 to +8192).
    *
    */
    void setPitchWheelValue(MidiEvent& evt, int value)
    {
        int midiValue=value+64*128;
        evt.byte1=midiValue&0x7F;
        evt.byte2=(midiValue/128)&0x7F;
    }

    /** For a channel after touch event, gets the after touch value (0-127)
    *
    */
    int  getChannelAfterTouchValue(const MidiEvent&  evt)
    {
        return evt.byte1&0x7F;
    }

    /** For a channel after touch event, sets the after touch value (0-127)
    *
    */
    void  setChannelAfterTouchValue(MidiEvent& evt, int value)
    {
        evt.byte1=(value&0x7F);;
    }
}
